//-----------------------------------------------------------------------------
// Copyright (C) Quarkslab. See README.md for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the Apache License as published by
// the Apache Software Foundation, either version 2.0 of the License.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See LICENSE.txt for the text of the Apache license.
//-----------------------------------------------------------------------------

#ifndef UTILS_HPP
#define UTILS_HPP

#include <array>
#include <cassert>
#include <cstdint>

#define LEN_ARRAY 256
#define AES_ROUND 14

#define MAYBE_UNUSED __attribute__((unused))

#endif /* UTILS_H */

